package com.tenlionsoft.baselib.widget.chat

import android.graphics.Bitmap
import android.graphics.BitmapShader
import android.graphics.Canvas
import android.graphics.Color
import android.graphics.ColorFilter
import android.graphics.Matrix
import android.graphics.Paint
import android.graphics.Path
import android.graphics.PixelFormat
import android.graphics.RectF
import android.graphics.Shader
import android.graphics.drawable.Drawable
import kotlin.math.min

class BubbleDrawable : Drawable {
    private var mRect: RectF? = null
    private val mPath = Path()
    private var mBitmapShader: BitmapShader? = null
    private val mPaint = Paint(Paint.ANTI_ALIAS_FLAG)
    private var mArrowWidth = 0f
    private var mAngle = 0f
    private var mArrowHeight = 0f
    private var mArrowPosition = 0f
    private var bubbleColor = 0
    private var bubbleBitmap: Bitmap? = null
    private var mArrowLocation: ArrowLocation? = null
    private var bubbleType: BubbleType? = null

    constructor(builder: Builder) {
        this.mRect = builder.mRect
        this.mAngle = builder.mAngle
        this.mArrowHeight = builder.mArrowHeight
        this.mArrowWidth = builder.mArrowWidth
        this.mArrowPosition = builder.mArrowPosition
        this.bubbleColor = builder.bubbleColor
        this.bubbleBitmap = builder.bubbleBitmap
        this.mArrowLocation = builder.mArrowLocation
        this.bubbleType = builder.bubbleType
    }


    override fun draw(canvas: Canvas) {
        setUp(canvas)
    }

    override fun getOpacity(): Int {
        return PixelFormat.TRANSLUCENT
    }

    override fun setAlpha(alpha: Int) {
        mPaint.alpha = alpha
    }

    override fun setColorFilter(cf: ColorFilter?) {
        mPaint.setColorFilter(cf)
    }

    private fun setUpPath(mArrowLocation: ArrowLocation, path: Path) {
        when (mArrowLocation) {
            ArrowLocation.LEFT -> setUpLeftPath(mRect!!, path)
            ArrowLocation.RIGHT -> setUpRightPath(mRect!!, path)
            ArrowLocation.TOP -> setUpTopPath(mRect!!, path)
            ArrowLocation.BOTTOM -> setUpBottomPath(mRect!!, path)
        }
    }

    private fun setUp(canvas: Canvas) {
        when (bubbleType!!) {
            BubbleType.COLOR -> mPaint.color = bubbleColor
            BubbleType.BITMAP -> {
                if (bubbleBitmap == null) return
                if (mBitmapShader == null) {
                    mBitmapShader = BitmapShader(
                        bubbleBitmap!!,
                        Shader.TileMode.CLAMP, Shader.TileMode.CLAMP
                    )
                }
                mPaint.setShader(mBitmapShader)
                setUpShaderMatrix()
            }
        }
        setUpPath(mArrowLocation!!, mPath)
        canvas.drawPath(mPath, mPaint)
    }

    private fun setUpLeftPath(rect: RectF, path: Path) {
        path.moveTo(mArrowWidth + rect.left + mAngle, rect.top)
        path.lineTo(rect.width() - mAngle, rect.top)
        path.arcTo(
            RectF(
                rect.right - mAngle, rect.top, rect.right,
                mAngle + rect.top
            ), 270f, 90f
        )
        path.lineTo(rect.right, rect.bottom - mAngle)
        path.arcTo(
            RectF(
                rect.right - mAngle, rect.bottom - mAngle,
                rect.right, rect.bottom
            ), 0f, 90f
        )
        path.lineTo(rect.left + mArrowWidth + mAngle, rect.bottom)
        path.arcTo(
            RectF(
                rect.left + mArrowWidth, rect.bottom - mAngle,
                mAngle + rect.left + mArrowWidth, rect.bottom
            ), 90f, 90f
        )
        path.lineTo(rect.left + mArrowWidth, mArrowHeight + mArrowPosition)
        path.lineTo(rect.left, mArrowPosition + mArrowHeight / 2)
        path.lineTo(rect.left + mArrowWidth, mArrowPosition)
        path.lineTo(rect.left + mArrowWidth, rect.top + mAngle)
        path.arcTo(
            RectF(
                rect.left + mArrowWidth, rect.top, mAngle
                        + rect.left + mArrowWidth, mAngle + rect.top
            ), 180f, 90f
        )
        path.close()
    }

    private fun setUpTopPath(rect: RectF, path: Path) {
        path.moveTo(
            (rect.left + min(mArrowPosition.toDouble(), mAngle.toDouble())).toFloat(),
            rect.top + mArrowHeight
        )
        path.lineTo(rect.left + mArrowPosition, rect.top + mArrowHeight)
        path.lineTo(rect.left + mArrowWidth / 2 + mArrowPosition, rect.top)
        path.lineTo(rect.left + mArrowWidth + mArrowPosition, rect.top + mArrowHeight)
        path.lineTo(rect.right - mAngle, rect.top + mArrowHeight)

        path.arcTo(
            RectF(
                rect.right - mAngle,
                rect.top + mArrowHeight, rect.right, mAngle + rect.top + mArrowHeight
            ), 270f, 90f
        )
        path.lineTo(rect.right, rect.bottom - mAngle)

        path.arcTo(
            RectF(
                rect.right - mAngle, rect.bottom - mAngle,
                rect.right, rect.bottom
            ), 0f, 90f
        )
        path.lineTo(rect.left + mAngle, rect.bottom)

        path.arcTo(
            RectF(
                rect.left, rect.bottom - mAngle,
                mAngle + rect.left, rect.bottom
            ), 90f, 90f
        )
        path.lineTo(rect.left, rect.top + mArrowHeight + mAngle)
        path.arcTo(
            RectF(
                rect.left, rect.top + mArrowHeight, mAngle
                        + rect.left, mAngle + rect.top + mArrowHeight
            ), 180f, 90f
        )
        path.close()
    }

    private fun setUpRightPath(rect: RectF, path: Path) {
        path.moveTo(rect.left + mAngle, rect.top)
        path.lineTo(rect.width() - mAngle - mArrowWidth, rect.top)
        path.arcTo(
            RectF(
                rect.right - mAngle - mArrowWidth,
                rect.top, rect.right - mArrowWidth, mAngle + rect.top
            ), 270f, 90f
        )
        path.lineTo(rect.right - mArrowWidth, mArrowPosition)
        path.lineTo(rect.right, mArrowPosition + mArrowHeight / 2)
        path.lineTo(rect.right - mArrowWidth, mArrowPosition + mArrowHeight)
        path.lineTo(rect.right - mArrowWidth, rect.bottom - mAngle)

        path.arcTo(
            RectF(
                rect.right - mAngle - mArrowWidth, rect.bottom - mAngle,
                rect.right - mArrowWidth, rect.bottom
            ), 0f, 90f
        )
        path.lineTo(rect.left + mArrowWidth, rect.bottom)

        path.arcTo(
            RectF(
                rect.left, rect.bottom - mAngle,
                mAngle + rect.left, rect.bottom
            ), 90f, 90f
        )

        path.arcTo(
            RectF(
                rect.left, rect.top, mAngle
                        + rect.left, mAngle + rect.top
            ), 180f, 90f
        )
        path.close()
    }

    private fun setUpBottomPath(rect: RectF, path: Path) {
        path.moveTo(rect.left + mAngle, rect.top)
        path.lineTo(rect.width() - mAngle, rect.top)
        path.arcTo(
            RectF(
                rect.right - mAngle,
                rect.top, rect.right, mAngle + rect.top
            ), 270f, 90f
        )

        path.lineTo(rect.right, rect.bottom - mArrowHeight - mAngle)
        path.arcTo(
            RectF(
                rect.right - mAngle, rect.bottom - mAngle - mArrowHeight,
                rect.right, rect.bottom - mArrowHeight
            ), 0f, 90f
        )

        path.lineTo(rect.left + mArrowWidth + mArrowPosition, rect.bottom - mArrowHeight)
        path.lineTo(rect.left + mArrowPosition + mArrowWidth / 2, rect.bottom)
        path.lineTo(rect.left + mArrowPosition, rect.bottom - mArrowHeight)
        path.lineTo(
            (rect.left + min(mAngle.toDouble(), mArrowPosition.toDouble())).toFloat(),
            rect.bottom - mArrowHeight
        )

        path.arcTo(
            RectF(
                rect.left, rect.bottom - mAngle - mArrowHeight,
                mAngle + rect.left, rect.bottom - mArrowHeight
            ), 90f, 90f
        )
        path.lineTo(rect.left, rect.top + mAngle)
        path.arcTo(
            RectF(
                rect.left, rect.top, mAngle
                        + rect.left, mAngle + rect.top
            ), 180f, 90f
        )
        path.close()
    }

    private fun setUpShaderMatrix() {
        val scale: Float
        val mShaderMatrix = Matrix()
        mShaderMatrix.set(null)
        val mBitmapWidth = bubbleBitmap!!.width
        val mBitmapHeight = bubbleBitmap!!.height
        val scaleX = intrinsicWidth / mBitmapWidth.toFloat()
        val scaleY = intrinsicHeight / mBitmapHeight.toFloat()
        scale = min(scaleX.toDouble(), scaleY.toDouble()).toFloat()
        mShaderMatrix.postScale(scale, scale)
        mShaderMatrix.postTranslate(mRect!!.left, mRect!!.top)
        mBitmapShader!!.setLocalMatrix(mShaderMatrix)
    }

    override fun getIntrinsicWidth(): Int {
        return mRect!!.width().toInt()
    }

    override fun getIntrinsicHeight(): Int {
        return mRect!!.height().toInt()
    }

    class Builder {
        internal var mRect: RectF? = null
        var mArrowWidth = DEFAULT_ARROW_WITH
        var mAngle = DEFAULT_ANGLE
        var mArrowHeight = DEFAULT_ARROW_HEIGHT
        var mArrowPosition = DEFAULT_ARROW_POSITION
        var bubbleColor = DEFAULT_BUBBLE_COLOR
        var bubbleBitmap: Bitmap? = null
        var bubbleType = BubbleType.COLOR
        var mArrowLocation = ArrowLocation.LEFT

        fun rect(rect: RectF?): Builder {
            this.mRect = rect
            return this
        }

        fun arrowWidth(mArrowWidth: Float): Builder {
            this.mArrowWidth = mArrowWidth
            return this
        }

        fun angle(mAngle: Float): Builder {
            this.mAngle = mAngle * 2
            return this
        }

        fun arrowHeight(mArrowHeight: Float): Builder {
            this.mArrowHeight = mArrowHeight
            return this
        }

        fun arrowPosition(mArrowPosition: Float): Builder {
            this.mArrowPosition = mArrowPosition
            return this
        }

        fun bubbleColor(bubbleColor: Int): Builder {
            this.bubbleColor = bubbleColor
            bubbleType(BubbleType.COLOR)
            return this
        }

        fun bubbleBitmap(bubbleBitmap: Bitmap?): Builder {
            this.bubbleBitmap = bubbleBitmap
            bubbleType(BubbleType.BITMAP)
            return this
        }

        fun arrowLocation(arrowLocation: ArrowLocation): Builder {
            this.mArrowLocation = arrowLocation
            return this
        }

        fun bubbleType(bubbleType: BubbleType): Builder {
            this.bubbleType = bubbleType
            return this
        }

        fun build(): BubbleDrawable {
            requireNotNull(mRect) { "BubbleDrawable Rect can not be null" }
            return BubbleDrawable(this)
        }

        companion object {
            var DEFAULT_ARROW_WITH: Float = 25f
            var DEFAULT_ARROW_HEIGHT: Float = 25f
            var DEFAULT_ANGLE: Float = 20f
            var DEFAULT_ARROW_POSITION: Float = 50f
            var DEFAULT_BUBBLE_COLOR: Int = Color.RED
        }
    }

    enum class ArrowLocation(val intValue: Int) {
        LEFT(0x00),
        RIGHT(0x01),
        TOP(0x02),
        BOTTOM(0x03);

        companion object {
            fun mapIntToValue(stateInt: Int): ArrowLocation {
                for (value in entries) {
                    if (stateInt == value.intValue) {
                        return value
                    }
                }
                return default
            }

            val default: ArrowLocation
                get() = LEFT
        }
    }

    enum class BubbleType {
        COLOR,
        BITMAP
    }
}